from genericpath import isfile
import json
import logging
import time
from selenium import webdriver
from selenium.webdriver.chrome.options import Options
from selenium.webdriver.common.by import By
from selenium.common.exceptions import NoSuchElementException
from selenium.webdriver.common.desired_capabilities import DesiredCapabilities
import os
from urllib.parse import urlparse
from urllib.parse import parse_qs
import os.path
from pathlib import Path
import requests
from milla_tdf.config import TradingFrameWorkConfig
from sys import platform
from urllib import request
import zipfile


class FiinAuthToken():
    def check_driver_and_download():
        CHROME_DRIVER_PATH = TradingFrameWorkConfig.selenium_driver_path

        if(isfile(CHROME_DRIVER_PATH) == True):
            return

        #create directory
        driver_path = os.path.normpath(CHROME_DRIVER_PATH + os.sep + os.pardir)
        Path(driver_path).mkdir(parents=True, exist_ok=True)

        mapping_file = {
            'win32' : 'chromedriver_win32.zip',
            'darwin' : 'chromedriver_mac64_m1.zip'
        }
        file = mapping_file[platform]

        logging.info(f'downloading {TradingFrameWorkConfig.selenium_download_url}/{file}')

        request.urlretrieve(f'''{TradingFrameWorkConfig.selenium_download_url}/{file}''', f"{driver_path}/{file}")

        with zipfile.ZipFile(f"{driver_path}/{file}", 'r') as zip_ref:
            zip_ref.extractall(f"{driver_path}")


    def get_token_by_selenium() -> str:

        FiinAuthToken.check_driver_and_download()

        logging.info('Auth with fiin trade ....')
        CHROME_DRIVER_PATH = TradingFrameWorkConfig.selenium_driver_path
        USERNAME=TradingFrameWorkConfig.fiin_trade_username
        PASSWORD=TradingFrameWorkConfig.fiin_trade_password

        os.chmod(CHROME_DRIVER_PATH, 755)

        caps = DesiredCapabilities.CHROME
        caps['goog:loggingPrefs'] = {'performance': 'ALL'}

        options = Options()
        options.add_argument('--headless')
        options.add_argument('--disable-gpu')
        options.add_argument("--log-level=3")
        driver = webdriver.Chrome(CHROME_DRIVER_PATH, chrome_options=options,desired_capabilities=caps)

        driver.get('https://fiintrade.vn')

        for i in range(10):
            try:
                driver.find_element(By.XPATH,'//*[@id="root"]/div[6]/header/div/div/div[3]/div[1]/div[2]').click()
                break
            except NoSuchElementException:
                time.sleep(3)

        for i in range(10):
            try:
                driver.find_element(By.XPATH,'//*[@id="exampleInputEmail1"]').send_keys(USERNAME)
                driver.find_element(By.XPATH,'//*[@id="exampleInputPassword1"]').send_keys(PASSWORD)
                driver.find_element(By.XPATH,'//*[@id="home"]/form/fieldset/div[3]/button').click()
                break
            except NoSuchElementException:
                time.sleep(3)

        def process_browser_log_entry(entry):
            response = json.loads(entry['message'])['message']
            return response

        browser_log = driver.get_log('performance') 
        events = [process_browser_log_entry(entry) for entry in browser_log]
        events = [event for event in events if 'Network.response' in event['method']]

        f = open("test.txt", "a")

        location:str = None
        for event in events:
            f.writelines(json.dumps(event))
            f.writelines('')
            f.writelines('')


        for event in events:
            try:
                if(event['params']['headers'].get('Location') != None):
                    location = event['params']['headers'].get('Location')
                    
            except KeyError:
                pass
            
        token = location.split('access_token=')[1].split('&')[0]
        driver.quit()
        logging.info(f'fiin trade token {token}')
        TradingFrameWorkConfig.fiin_cookie = f'Bearer {token}'

        return token


from dataclasses import dataclass
from typing import Callable


@dataclass
class FieldMappingConfig:
    new_field_name:str
    function: Callable
    args:dict

def remap_data_value(kwargs):
    return kwargs['mapping_data'][kwargs['raw_value']]


from copy import copy
from datetime import datetime
import json
import logging
from typing import Union

import pandas as pd
from milla_tdf.config import TradingFrameWorkConfig
from milla_tdf.crawler.abstract import AbstractCrawler, UpdateModeEnum
from milla_tdf.data_model.abstract import AbstractDataModel, AbstractMetaData
from milla_tdf.data_model.stock_price_data import StockPriceData, StockPriceMetaData


class FiinStockPriceDataCrawler(AbstractCrawler):
    data_model = StockPriceData
    update_mode = UpdateModeEnum.APPEND
    @classmethod
    def crawl_data(cls, meta_data: AbstractMetaData,**kwargs) -> list[StockPriceData]:
        super().crawl_data(data_model=cls.data_model,meta_data=meta_data)
        symbols = kwargs['symbols']

        result: list[StockPriceData] = []

        for symbol in symbols:
            latest_update_date = cls.get_latest_update_date(meta_data,symbol)

            if(latest_update_date != None):
                cls.call_api_and_append_result(symbol,latest_update_date,datetime.now().strftime('%Y-%m-%d'),result)

        return result

    @classmethod
    def call_api_and_append_result(cls,symbol:str,start:str,end:str, result:list[StockPriceData],**kwargs):
        date_range_arr = cls.split_calling_arr(start,end)

        for date_range in date_range_arr:
            logging.info(f'''calling data {symbol} from {date_range['start']} -> {date_range['end']}''')

            raw_data = json.loads(cls.call_fiin_api_get(f'{TradingFrameWorkConfig.fiin_technical_end_point}/TradingView/GetStockChartData',
                {'language':'vi','Code':symbol,'Frequency':'Daily',
                    'From':date_range['start'],
                    'To':date_range['end'],
                    'Type':'Stock'
                }
            ))['items']

            for data in raw_data:
                result.append(StockPriceData(symbol,data['tradingDate'],data['openPrice'],data['highestPrice'],data['lowestPrice'],data['closePrice'],data['totalMatchVolume']))
        
    @classmethod
    def split_calling_arr(cls,start:str,end:str):
        start = datetime.strptime(start,'%Y-%m-%d')
        end = datetime.strptime(end,'%Y-%m-%d')

        result = []

        while start < end:
            result.append({'start':start.strftime('%Y-%m-%dT%H:%M:%SZ'),'end':copy(start).replace(year=(start.year) + 1).strftime('%Y-%m-%dT%H:%M:%SZ')})
            start = start.replace(year=(start.year) + 1)

        return result

    @classmethod
    def get_latest_update_date(cls,old_meta_data:StockPriceMetaData,symbol:str, **kwargs)  -> Union[str,None]:
        if(old_meta_data == None):
            return TradingFrameWorkConfig.data_start_at
        try:
            updated = old_meta_data.updated[old_meta_data.symbols.index(symbol)] 

            if(updated == datetime.now().strftime('%Y-%m-%d') ):
                return None
            else:
                return updated
        except ValueError:
            pass

        return TradingFrameWorkConfig.data_start_at

    @classmethod
    def build_meta_data(cls,old_meta_data:StockPriceMetaData, **kwargs)  -> AbstractMetaData:
        super().build_meta_data(old_meta_data,**kwargs)
        if(old_meta_data == None):
            #build new meta data
            symbols = kwargs['symbols']
            updated = [datetime.now().strftime('%Y-%m-%d') for x in symbols]
            return StockPriceMetaData(symbols,updated)
        else:
            #update old metadata
            return cls.find_and_update_old_metadata(old_meta_data,**kwargs)

    
    @classmethod
    def find_and_update_old_metadata(cls,old_meta_data:StockPriceMetaData, **kwargs) -> StockPriceMetaData:
        old_symbols = old_meta_data.symbols
        old_updated = old_meta_data.updated
        updated_symbols = kwargs['symbols']
        new_symbols = []
        update_indexes = []
        for symbol in updated_symbols:
            try:
                update_indexes.append(old_symbols.index(symbol))
            except ValueError:
                new_symbols.append(symbol)
        
        if(len(update_indexes) != 0):
            for update_index in update_indexes:
                old_updated[update_index] = datetime.now().strftime('%Y-%m-%d')

        if(len(new_symbols) != 0):
            for new_symbol in new_symbols:
                old_symbols.append(new_symbol)
                old_updated.append(datetime.now().strftime('%Y-%m-%d'))

        return StockPriceMetaData(old_symbols,old_updated)

    @classmethod
    def filter_selected_df(cls,df:pd.DataFrame, **kwargs) -> pd.DataFrame:
        logging.info('filter')
        symbols = kwargs['symbols']

        return df[df['Symbol'].isin(symbols) ]




from dataclasses import dataclass

from dataclasses_json import dataclass_json

from milla_tdf.data_model.abstract import AbstractDataModel


@dataclass
class SymbolInfoData(AbstractDataModel):
    Symbol:str
    OrganCode:str
    Board:str
    ComTypeCode:str
    
    


@dataclass_json
@dataclass
class SymbolInfoMetaData():
    updated:str


from copy import copy
import logging
from typing import Callable, Union

import pandas as pd
from milla_tdf.config import TradingFrameWorkConfig
from milla_tdf.crawler.error_codes import CrawlerError
from milla_tdf.data_model.abstract import AbstractDataModel, AbstractMetaData
from enum import Enum
import requests
from dacite import from_dict
from milla_tdf.field_mapping import FieldMappingConfig

class UpdateModeEnum(Enum):
     APPEND = 'APPEND'
     OVERWRITE = 'OVERWRITE'

class AbstractCrawler():
    data_model:AbstractDataModel
    update_mode:UpdateModeEnum
    @classmethod
    def crawl_data(cls, meta_data: AbstractMetaData, **kwargs) -> list[AbstractDataModel]:
        logging.debug(
            f'{cls.data_model.__name__} -> crawl data_model {cls.data_model.__name__}')
    
    @classmethod
    def build_meta_data(cls,old_meta_data:AbstractMetaData, **kwargs)  -> AbstractMetaData:
        pass

    @classmethod
    def call_fiin_api_get(cls,url:str,params:dict) -> str:

        headers = {
            'Accept': 'application/json, text/plain, */*',
            'Accept-Language': 'en-US,en;q=0.9',
            'Authorization': TradingFrameWorkConfig.fiin_cookie,
            'Connection': 'keep-alive',
            'Origin': 'https://fiintrade.vn',
            'Referer': 'https://fiintrade.vn/',
            'Sec-Fetch-Dest': 'empty',
            'Sec-Fetch-Mode': 'cors',
            'Sec-Fetch-Site': 'same-site',
            'User-Agent': 'Mozilla/5.0 (Macintosh; Intel Mac OS X 10_15_7) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/103.0.0.0 Safari/537.36',
            'sec-ch-ua': '".Not/A)Brand";v="99", "Google Chrome";v="103", "Chromium";v="103"',
            'sec-ch-ua-mobile': '?0',
            'sec-ch-ua-platform': '"macOS"'
        }

        response = requests.request("GET", url, headers=headers,params=params)
        
        if(response.status_code == 401):
            logging.error(f'''Please recheck Fiin's session, username, password''')
            raise CrawlerError(f'Unauthorized')

        return response.text

    @classmethod
    def filter_selected_df(cls,df:pd.DataFrame, **kwargs):
        #by default
        return df 
        

    @classmethod
    def construct_data_model_arr(cls,raw_data_arr:list[dict], data_model:AbstractDataModel, mapping_fields:dict) -> list[AbstractDataModel]:
        data_arr = []

        for raw_data in raw_data_arr:
            data ={}
            for mapping_field in mapping_fields.items():
                key = mapping_field[0]
                value = mapping_field[1]

                if(value.__class__ == str):
                    data[value] = raw_data[key]
                elif(value.__class__ == FieldMappingConfig):
                    function:Callable = value.function
                    args:dict = value.args
                    new_field_name:str = value.new_field_name
                    new_args = copy(args)
                    new_args['raw_value'] = raw_data[key]
                    data[new_field_name] = function(new_args)

            data = from_dict(data_class=data_model, data=data)
            data_arr.append(data)

        return data_arr

                




from copy import copy
from datetime import datetime
import json
import logging
from dacite import Config, from_dict
from dateutil.relativedelta import relativedelta
import pandas as pd
from milla_tdf.config import TradingFrameWorkConfig
from milla_tdf.crawler.abstract import AbstractCrawler, UpdateModeEnum
from milla_tdf.data_model.abstract import AbstractDataModel, AbstractMetaData
from milla_tdf.data_model.com_financial_data import ComFinancialData, ComFinancialMetaData
from milla_tdf.data_model.stock_price_data import StockPriceData, StockPriceMetaData

class FiinComFinancialDataCrawler(AbstractCrawler):
    data_model = ComFinancialData
    update_mode = UpdateModeEnum.APPEND
    @classmethod
    def crawl_data(cls, meta_data: ComFinancialMetaData,**kwargs) -> list[StockPriceData]:
        symbols = kwargs['symbols']

        result: list[ComFinancialData] = []
        
        symbol_info_df:pd.DataFrame = kwargs['symbol_info_df']

        for symbol in symbols:
            time_lines = cls.build_timeline_arr(symbol,meta_data,**kwargs)

            if(time_lines == None):
                continue

            logging.info(f'crawl {cls.data_model.__name__} symbol {symbol} timelines {time_lines}')

            organ_code = symbol_info_df[symbol_info_df['Symbol'] == symbol]['OrganCode'].iloc[0]

            logging.debug(f'''symbol {symbol} --> organ_code {organ_code}''')

            url = f'{TradingFrameWorkConfig.fiin_fundamental_end_point}/FinancialAnalysis/GetFinancialRatioV2?language=vi&Type=Company&OrganCode={organ_code}'

            for time_line in time_lines:
                url = f'{url}&Timeline={time_line}'

            response = cls.call_fiin_api_get(url,{})

            raw_data = json.loads(response)['items']

            for data in raw_data:
                if(data['value'] == None):
                    continue
                
                if(data['value']['organCode'] != 'EndOfData'):
                    data['value']['Symbol'] = symbol
                    result.append(from_dict(data_class=ComFinancialData, data=data['value'],config=Config(check_types=False)))

        return result

    @classmethod
    def build_timeline_arr(cls,symbol,old_meta_data:ComFinancialMetaData, **kwargs) -> list[str]:
        try:
            latest_updated = old_meta_data.updated[old_meta_data.symbols.index(symbol)]
        except ValueError :
            latest_updated = TradingFrameWorkConfig.data_start_at
        except AttributeError:
            latest_updated = TradingFrameWorkConfig.data_start_at

        
        if(latest_updated == datetime.now().strftime('%Y-%m-%d')):
            return None
        
        start = datetime.strptime(latest_updated,'%Y-%m-%d')
        end = datetime.now()

        result = []
        while start < end:
            result.append(f'{start.year}_{pd.Timestamp(start).quarter}')
            start = start + relativedelta(months=+3)

        return result

    @classmethod
    def build_meta_data(cls,old_meta_data:ComFinancialMetaData, **kwargs)  -> AbstractMetaData:
        super().build_meta_data(old_meta_data,**kwargs)
        if(old_meta_data == None):
            #build new meta data
            symbols = kwargs['symbols']
            updated = [datetime.now().strftime('%Y-%m-%d') for x in symbols]
            return ComFinancialMetaData(symbols,updated)
        else:
            #update old metadata
            return cls.find_and_update_old_metadata(old_meta_data,**kwargs)

    
    @classmethod
    def find_and_update_old_metadata(cls,old_meta_data:ComFinancialMetaData, **kwargs) -> StockPriceMetaData:
        old_symbols = old_meta_data.symbols
        old_updated = old_meta_data.updated
        updated_symbols = kwargs['symbols']
        new_symbols = []
        update_indexes = []
        for symbol in updated_symbols:
            try:
                update_indexes.append(old_symbols.index(symbol))
            except ValueError:
                new_symbols.append(symbol)
        
        if(len(update_indexes) != 0):
            for update_index in update_indexes:
                old_updated[update_index] = datetime.now().strftime('%Y-%m-%d')

        if(len(new_symbols) != 0):
            for new_symbol in new_symbols:
                old_symbols.append(new_symbol)
                old_updated.append(datetime.now().strftime('%Y-%m-%d'))

        return ComFinancialMetaData(old_symbols,old_updated)

    @classmethod
    def filter_selected_df(cls,df:pd.DataFrame, **kwargs) -> pd.DataFrame:
        symbols = kwargs['symbols']

        return df[df['Symbol'].isin(symbols) ]




from datetime import datetime
import logging

import pandas as pd
from milla_tdf.config import TradingFrameWorkConfig
from milla_tdf.crawler.abstract import AbstractCrawler, UpdateModeEnum
from milla_tdf.data_model.abstract import AbstractDataModel, AbstractMetaData
from milla_tdf.data_model.stock_price_data import StockPriceData, StockPriceMetaData
from milla_tdf.data_model.symbols_info import SymbolInfoData, SymbolInfoMetaData
from milla_tdf.field_mapping import FieldMappingConfig,remap_data_value
import json

class FiinSymbolsInfoCrawler(AbstractCrawler):
    data_model = SymbolInfoData
    update_mode = UpdateModeEnum.OVERWRITE
    @classmethod
    def crawl_data(cls, meta_data: SymbolInfoMetaData,**kwargs) -> list[SymbolInfoData]:
        super().crawl_data(data_model=cls.data_model,meta_data=meta_data)
        if(meta_data != None and meta_data.updated == datetime.now().strftime('%Y-%m-%d')):
            logging.info(f'skip crawling {cls.__name__}, loaded from local')
            return []

        raw_data = json.loads(cls.call_fiin_api_get(f'{TradingFrameWorkConfig.fiin_core_end_point}/Master/GetListOrganization',{'language':'vi'}))['items']

        data: list[SymbolInfoData]= cls.construct_data_model_arr(raw_data,SymbolInfoData,
            {
                'organCode':'OrganCode',
                'ticker':'Symbol',
                'comTypeCode':'ComTypeCode',
                'comGroupCode': FieldMappingConfig(
                    new_field_name='Board',
                    function=remap_data_value,
                    args = {
                        'mapping_data': {
                            'VNINDEX':'HSX',
                            'UpcomIndex':'UPCOM',
                            'HNXIndex':'HNX'
                        }
                    },
                )
            })

        return data

    @classmethod
    def build_meta_data(cls,old_meta_data:AbstractMetaData, **kwargs) -> SymbolInfoData:
        return SymbolInfoMetaData(updated=datetime.now().strftime('%Y-%m-%d'))

        




